@@ -1,4 +1,10 @@
-from util import gen_name, flattener
+# -*- coding: utf-8 -*-
+
+from __future__ import (absolute_import, division, print_function,
+                        unicode_literals)
+from .util import gen_name, flattener
+from nine import iteritems, nine
+
 
 def generate_python(ir):
     cur_indent = 0
@@ -10,13 +16,13 @@ def generate_python(ir):
         for line in node.py():
             yield line.indent(cur_indent)
 
-class Node(object):
 
+class Node(object):
     def __init__(self):
         self.filename = '<string>'
         self.lineno = 0
 
-    def py(self): # pragma no cover
+    def py(self):  # pragma no cover
         return []
 
     def __iter__(self):
@@ -25,8 +31,16 @@ class Node(object):
     def line(self, text):
         return PyLine(self.filename, self.lineno, text)
 
-class HierNode(Node):
 
+class PassNode(Node):
+    def py(self):
+        # 'pass' would result in: TypeError: 'NoneType' object is not iterable
+        yield self.line('yield ""')
+
+
+class HierNode(Node):
+    '''Base for nodes that contain an indented Python block (def, for, if etc.)
+    '''
     def __init__(self, body):
         super(HierNode, self).__init__()
         self.body = tuple(x for x in body if x is not None)
@@ -38,23 +52,31 @@ class HierNode(Node):
     def __iter__(self):
         yield self
         yield IndentNode()
-        for x in self.body_iter(): yield x
+        for x in self.body_iter():
+            yield x
         yield DedentNode()
 
-class IndentNode(Node): pass
-class DedentNode(Node): pass
+
+class IndentNode(Node):
+    pass
+
+
+class DedentNode(Node):
+    pass
+
 
 class TemplateNode(HierNode):
-
     class TemplateTail(Node):
         def py(self):
             yield self.line('template = kajiki.Template(template)')
 
     def __init__(self, mod_py=None, defs=None):
         super(TemplateNode, self).__init__(defs)
-        if mod_py is None: mod_py = []
-        if defs is None: defs = []
-        self.mod_py = [ x for x in mod_py if x is not None ]
+        if mod_py is None:
+            mod_py = []
+        if defs is None:
+            defs = []
+        self.mod_py = [x for x in mod_py if x is not None]
 
     def py(self):
         yield self.line('class template:')
@@ -66,8 +88,8 @@ class TemplateNode(HierNode):
             yield x
         yield self.TemplateTail()
 
-class ImportNode(Node):
 
+class ImportNode(Node):
     def __init__(self, tpl_name, alias=None):
         super(ImportNode, self).__init__()
         self.tpl_name = tpl_name
@@ -75,22 +97,22 @@ class ImportNode(Node):
 
     def py(self):
         yield self.line(
-            'local.__kj__.import_(%r, %r, globals())' % (
+            'local.__kj__.import_(%r, %r, self.__globals__)' % (
                 self.tpl_name, self.alias))
 
-class IncludeNode(Node):
 
+class IncludeNode(Node):
     def __init__(self, tpl_name):
         super(IncludeNode, self).__init__()
         self.tpl_name = tpl_name
 
     def py(self):
         yield self.line(
-            'yield local.__kj__.import_(%r, None, {}).__main__()' % (
+            'yield local.__kj__.import_(%r, None, self.__globals__).__main__()' % (
                 self.tpl_name))
 
-class ExtendNode(Node):
 
+class ExtendNode(Node):
     def __init__(self, tpl_name):
         super(ExtendNode, self).__init__()
         self.tpl_name = tpl_name
@@ -100,6 +122,7 @@ class ExtendNode(Node):
             'yield local.__kj__.extend(%r).__main__()' % (
                 self.tpl_name))
 
+
 class DefNode(HierNode):
     prefix = '@kajiki.expose'
 
@@ -111,17 +134,30 @@ class DefNode(HierNode):
         yield self.line(self.prefix)
         yield self.line('def %s:' % (self.decl))
 
+    def __iter__(self):
+        yield self
+        yield IndentNode()
+        is_empty = True
+        for x in self.body_iter():
+            yield x
+            is_empty = False
+        if is_empty:  # In Python, a function without a body is a SyntaxError.
+            yield PassNode()  # Prevent creation of a function without a body
+        yield DedentNode()
+
+
 class InnerDefNode(DefNode):
-    prefix='@__kj__.flattener.decorate'
+    prefix = '@__kj__.flattener.decorate'
+
 
 class CallNode(HierNode):
-
     class CallTail(Node):
         def __init__(self, call):
             super(CallNode.CallTail, self).__init__()
             self.call = call
+
         def py(self):
-            yield self.line('yield ' + self.call) 
+            yield self.line('yield ' + self.call)
 
     def __init__(self, caller, callee, *body):
         super(CallNode, self).__init__(body)
@@ -136,12 +172,13 @@ class CallNode(HierNode):
     def __iter__(self):
         yield self
         yield IndentNode()
-        for x in self.body_iter(): yield x
+        for x in self.body_iter():
+            yield x
         yield DedentNode()
         yield self.CallTail(self.call)
 
-class ForNode(HierNode):
 
+class ForNode(HierNode):
     def __init__(self, decl, *body):
         super(ForNode, self).__init__(body)
         self.decl = decl
@@ -149,12 +186,13 @@ class ForNode(HierNode):
     def py(self):
         yield self.line('for %s:' % (self.decl))
 
-class WithNode(HierNode):
 
+class WithNode(HierNode):
     class WithTail(Node):
         def __init__(self, vars):
             super(WithNode.WithTail, self).__init__()
             self.vars = vars
+
         def py(self):
             gen = gen_name()
             yield self.line('%s = local.__kj__.pop_with()' % gen)
@@ -171,17 +209,17 @@ class WithNode(HierNode):
     def py(self):
         yield self.line(
             'local.__kj__.push_with(locals(), %s)' % self.vars_text)
-        for k,v in self.vars.iteritems():
-            yield self.line('%s = %s' % (k,v))
-        
+        for k, v in iteritems(self.vars):
+            yield self.line('%s = %s' % (k, v))
+
     def __iter__(self):
         yield self
-        for x in self.body_iter(): yield x
+        for x in self.body_iter():
+            yield x
         yield self.WithTail(self.vars)
-    
+
 
 class SwitchNode(HierNode):
-
     class SwitchTail(Node):
         def py(self):
             yield self.line('local.__kj__.pop_switch()')
@@ -195,11 +233,12 @@ class SwitchNode(HierNode):
 
     def __iter__(self):
         yield self
-        for x in self.body_iter(): yield x
+        for x in self.body_iter():
+            yield x
         yield self.SwitchTail()
 
-class CaseNode(HierNode):
 
+class CaseNode(HierNode):
     def __init__(self, decl, *body):
         super(CaseNode, self).__init__(body)
         self.decl = decl
@@ -207,8 +246,8 @@ class CaseNode(HierNode):
     def py(self):
         yield self.line('if local.__kj__.case(%s):' % self.decl)
 
-class IfNode(HierNode):
 
+class IfNode(HierNode):
     def __init__(self, decl, *body):
         super(IfNode, self).__init__(body)
         self.decl = decl
@@ -216,16 +255,17 @@ class IfNode(HierNode):
     def py(self):
         yield self.line('if %s:' % self.decl)
 
-class ElseNode(HierNode):
 
+class ElseNode(HierNode):
     def __init__(self,  *body):
         super(ElseNode, self).__init__(body)
 
     def py(self):
         yield self.line('else:')
 
-class TextNode(Node):
 
+class TextNode(Node):
+    '''Node that outputs Python literals.'''
     def __init__(self, text, guard=None):
         super(TextNode, self).__init__()
         self.text = text
@@ -238,8 +278,8 @@ class TextNode(Node):
         else:
             yield self.line(s)
 
-class TranslatableTextNode(TextNode):
 
+class TranslatableTextNode(TextNode):
     def py(self):
         text = self.text.strip()
         if text:
@@ -251,8 +291,11 @@ class TranslatableTextNode(TextNode):
         else:
             yield self.line(s)
 
-class ExprNode(Node):
 
+class ExprNode(Node):
+    '''Node that contains a Python expression to be evaluated when the template
+    is executed.
+    '''
     def __init__(self, text, safe=False):
         super(ExprNode, self).__init__()
         self.text = text
@@ -264,17 +307,14 @@ class ExprNode(Node):
         else:
             yield self.line('yield self.__kj__.escape(%s)' % self.text)
 
-class PassNode(Node):
-
-    def py(self):
-        yield self.line('yield ""')
 
 class AttrNode(HierNode):
-
+    '''Node that renders HTML/XML attributes.'''
     class AttrTail(Node):
         def __init__(self, parent):
             super(AttrNode.AttrTail, self).__init__()
             self.p = parent
+
         def py(self):
             gen = self.p.genname
             x = gen_name()
@@ -287,10 +327,9 @@ class AttrNode(HierNode):
     def __init__(self, attr, value, guard=None, mode='xml'):
         super(AttrNode, self).__init__(value)
         self.attr = attr
-        # self.value = value
         self.guard = guard
         self.mode = mode
-        self.attrname, self.genname = gen_name(), gen_name()
+        self.genname = gen_name()
 
     def py(self):
         yield self.line('def %s():' % self.genname)
@@ -313,8 +352,8 @@ class AttrNode(HierNode):
             yield DedentNode()
             yield self.AttrTail(self)
 
-class AttrsNode(Node):
 
+class AttrsNode(Node):
     def __init__(self, attrs, guard=None, mode='xml'):
         super(AttrsNode, self).__init__()
         self.attrs = attrs
@@ -323,19 +362,22 @@ class AttrsNode(Node):
 
     def py(self):
         x = gen_name()
+
         def _body():
             yield self.line(
-                'for %s in self.__kj__.render_attrs(%s, %r):' % (x, self.attrs, self.mode))
+                'for %s in self.__kj__.render_attrs(%s, %r):' % (
+                    x, self.attrs, self.mode))
             yield self.line('    yield %s' % x)
         if self.guard:
             yield self.line('if %s:' % self.guard)
             for l in _body():
                 yield l.indent()
         else:
-            for l in _body(): yield l
+            for l in _body():
+                yield l
+
 
 class PythonNode(Node):
-
     def __init__(self, *body):
         super(PythonNode, self).__init__()
         self.module_level = False
@@ -360,28 +402,32 @@ class PythonNode(Node):
         for line in text.splitlines():
             if prefix is None:
                 rest = line.lstrip()
-                prefix = line[:len(line)-len(rest)]
+                prefix = line[:len(line) - len(rest)]
             assert line.startswith(prefix)
             yield line[len(prefix):]
 
+
 def optimize(iter_node):
     last_node = None
     for node in iter_node:
         if type(node) == TextNode:
             if (type(last_node) == TextNode
-                and last_node.guard == node.guard):
+                    and last_node.guard == node.guard):
                 last_node.text += node.text
             else:
-                if last_node is not None: yield last_node
+                if last_node is not None:
+                    yield last_node
                 last_node = node
         else:
-            if last_node is not None: yield last_node
+            if last_node is not None:
+                yield last_node
             last_node = node
     if last_node is not None:
         yield last_node
 
-class PyLine(object):
 
+@nine
+class PyLine(object):
     def __init__(self, filename, lineno, text, indent=0):
         self._filename = filename
         self._lineno = lineno
@@ -389,15 +435,16 @@ class PyLine(object):
         self._indent = indent
 
     def indent(self, sz=4):
-        return PyLine(self._filename, self._lineno, self._text, self._indent + sz)
+        return PyLine(self._filename, self._lineno, self._text,
+                      self._indent + sz)
 
     def __str__(self):
         return (' ' * self._indent) + self._text
         if self._lineno:
-            return (' ' * self._indent) + self._text + '\t# %s:%d' % (self._filename, self._lineno)
+            return (' ' * self._indent) + self._text + '\t# %s:%d' % (
+                self._filename, self._lineno)
         else:
             return (' ' * self._indent) + self._text
 
     def __repr__(self):
         return '%s:%s %s' % (self._filename, self._lineno, self)
-
